/*
*	ftrim trims file giving starting and ending point.
*
*	you might be tempted to use "dd":
*	dd -if $target -ibs 1 -iseek $size -quiet 1
*	but dd is very slow in network environment
*
*	coded by
*	Kenar (Kenji Arisawa)
*	E-mail: arisawa@aichi-u.ac.jp
*/

#include <u.h>
#include <libc.h>
#include <ctype.h>
#include "misc.h"

#define BUFSIZE 16*1024
#define MAXFIELD 3

enum
{
	KB = 1024,
	MB = 1024*KB,
	GB = 1024*MB,
};

void
usage(void)
{
	fprint(2,
	"usage: ftrim [-s $start] [-e $end] [-f $range] $file\n"
	"  -s $start	# starting address\n"
	"  -e $end		# ending address\n"
	"  -f $range		# a range file. the line format is \"start end\" in unit of byte\n"
	"if none of those options is given, the range must come from stdin\n"
	);
	exits("usage");
}

void
getunit(char *buf,int size, char *str)
{
	char *s;
	s = str;
	while(isdigit(*s++));
	snprint(buf,size,"%s",s);
}

void
spreduce(char *str)
/*	remove redundant space and tabs in the string	*/
{
	char *sep = " 	"; // space and TAB	
	int i,j,n;
	char ch, *s;
	i = 0;
	j = 1;
	n = strlen(str);
	for(s = str; *s; s++)
		if(strchr(sep,*s))
			*s = ' ';
	while(j < n){
		ch = str[j];
		if(ch == ' ' && str[i] == ' ')
			j++;
		else{
			i++;
			str[i] = ch;
			j++;
		}
	}
	if(str[i] == ' ')
		str[i] = 0;
	str[i+1] = 0;
}

void
seekread(int fd, vlong sp, vlong ep)
{
	vlong cp,n;
	char buf[BUFSIZE];

	if(seek(fd, sp, 0) < 0)
		sysfatal("fseek: %r");
	cp = sp;
	while(BUFSIZE < ep -cp && (n = read(fd, buf, BUFSIZE)) > 0){
		write(1, buf, n);
		cp += n;
	}
	n = read(fd, buf, ep - cp);
	write(1, buf, n);
}

void main(int argc, char *argv[])
{	char *a, *p, *args[MAXFIELD];
	char buf[10];
	vlong sp = -1;
	vlong ep = -1;
	vlong ep0;
	char *rangefile = nil;
	int m, fd;
	int fd1 = 0;
	Dir *dir;
	Rdbuf *rb;

	ARGBEGIN{
	case 's':	a=ARGF();
		if(a == nil) usage();
		sp = atoll(a);
		getunit(buf,3,a);
		if(strcmp(buf,"KB") == 0)
			sp *= KB;
		else if(strcmp(buf,"MB") == 0)
			sp *= MB;
		else if(strcmp(buf,"GB") == 0)
			sp *= GB;
		break;
	case 'e':	a=ARGF();
		if(a == nil) usage();
		ep = atoll(a);
		getunit(buf,3,a);
		if(strcmp(buf,"KB") == 0)
			ep *= KB;
		else if(strcmp(buf,"MB") == 0)
			ep *= MB;
		else if(strcmp(buf,"GB") == 0)
			ep *= GB;
		break;
	case 'f': rangefile = EARGF(usage());
		break;
	default: usage();
	}ARGEND

	if(*argv == nil)
		usage();
	fd = open(*argv, OREAD);
	if(fd<0)
		usage();
	dir = dirfstat(fd);
	ep0 = dir->length;

	if(sp == -1 && ep == -1 && rangefile == nil)
		goto L0;

	if(rangefile == nil){
		if(sp == -1)
			sp = 0;
		if(ep == -1){
			dir = dirfstat(fd);
			ep = dir->length;
		}
		seekread(fd,sp,ep);
		close(fd);
		exits(nil);
	}

	fd1 = open(rangefile, OREAD);
	if(fd1 < 0)
		sysfatal("%s not open: %r", rangefile);

L0:
	rb = rdinit(fd1, 32*1024);
	while((p = rdline(rb))){ /* assign = */
		spreduce(p);
		m = split(p, args, MAXFIELD, ' ');
		if(m > 2 || m == 0)
			sysfatal("invalid format");
		sp = atoll(args[0]);
		if(m == 1)
			ep = ep0;
		else
			ep = atoll(args[1]);
		seekread(fd,sp,ep);
	}
	rdfree(rb);
	if(fd1 != 0)
		close(fd1);
	close(fd);
	/*	without exits(nil) fseek returns error status
	*	man seek(2): "Sets errstr." */
	exits(nil);
}
